#pragma once
#include <cstdint>
#include <vector>
#include <Event.h>
#include <map>
#include <thread>
#include <ReWindow/types/Key.h>
#include <ReWindow/types/Cursors.h>
#include <ReWindow/types/MouseButton.h>
#include <ReWindow/types/GamepadButton.h>
#include <ReWindow/types/Pair.h>
#include <ReWindow/types/WindowEvents.h>
#include <queue>


namespace ReWindow {
    struct KeyboardState { std::map<Key, bool> PressedKeys; };
    struct GamepadState { std::map<GamepadButton, bool> PressedButtons; };

    class MouseState;
    class RWindow;
    class OpenGLWindow;
    class VulkanWindow;
}

enum class WindowFlag: uint8_t {
    IN_FOCUS, FULLSCREEN, RESIZABLE,
    VSYNC, QUIT, MAX_FLAG
};
std::string WindowFlagToStr(WindowFlag flag);

class ReWindow::MouseState {
public:
    struct
    {
        bool LMB = false;
        bool RMB = false;
        bool MMB = false;
        bool SideButton1 = false;
        bool SideButton2 = false;
        bool MWheelUp = false;
        bool MWheelDown = false;
    }

    Buttons;
    IPair Position;
    int Wheel = 0;
public:
    [[nodiscard]] bool IsDown(const MouseButton& btn) const;
    void Set(const MouseButton& btn, bool state);
    bool& operator[](const MouseButton& btn);
};

/// RWindow is a class implementation of a platform-independent window abstraction.
/// This library also provides abstractions for user-input devices, and their interaction with the window.
class ReWindow::RWindow {
protected:
    class Platform;
    Platform* platform;
    inline static RWindow* extant;

    int width = 1280;
    int height = 720;

    bool open = false; // Is the underlying OS-Window-Handle actually open.
    bool resizable = true;
    bool fullscreen_mode = false;
    bool focused = true;
    bool vsync = false;
    bool cursor_visible = true;
    bool closing = false;

    float delta_time = 0.f;
    float refresh_rate = 0.f;
    unsigned int refresh_count = 0;
    IPair render_area_position = {0, 0};
    std::string title = "Redacted Window";

    IPair lastKnownWindowSize {0, 0};
    // TODO remove the flags.
    bool flags[5];
    std::vector<RWindowEvent> eventLog;	// history of all logged window events.
    std::queue<RWindowEvent> eventQueue; //
    MouseState currentMouse;		// purrent frame mouse state.
    MouseState previousMouse;		// previous frame mouse state

    // TODO: Why tf are these on the header and not just in the cpp?
    float refresh_rate_prev_1 = 0.f;
    float refresh_rate_prev_2 = 0.f;
    float refresh_rate_prev_3 = 0.f;
    float refresh_rate_prev_4 = 0.f;
    float refresh_rate_prev_5 = 0.f;
    float avg_refresh_rate = 0.0f;

    /// Returns the most accurate and recent available mouse coordinates.
    /// @note Call this version at most **once** per-frame. It polls the X-Window server and therefore is quite slow.
    /// @see getCursorPos();
    [[nodiscard]] IPair GetAccurateMouseCoordinates() const;
protected:
    void LogEvent(const RWindowEvent& e) { eventLog.push_back(e);}
    //void EnqueueEvent(const RWindowEvent& e) { eventQueue.push(e);}
    [[nodiscard]] RWindowEvent GetLastEvent() const { return eventLog.back(); }
    /// Requests the operating system to make the window fullscreen. Saves the previous window size as well.
    void Fullscreen();
    /// Requests the operating system to take the window out of fullscreen mode. Previously saved window size is restored, if possible.
    void RestoreFromFullscreen();

public:
    KeyboardState currentKeyboard;  // current frame keyboard state.
    KeyboardState previousKeyboard; // previous frame keyboard state.
    // TODO: Josh hates parameter-flags, it's not 1995 :/
    [[nodiscard]] bool GetFlag(WindowFlag flag) const;
    void SetFlag(WindowFlag flag, bool state);

    /// The default constructor does not set any members, and are left uninitialized.
    RWindow();
    /// Constructs a window by explicitly setting the title, width, height, and optionally; rendering API, fullscreen, resizable, and vsync.
    /// @param wTitle The window title text.
    /// @param wWidth
    /// @param wHeight
    /// @param wFullscreen
    /// @param wResizable
    /// @param wVsync
    explicit RWindow(const std::string& wTitle, int wWidth = 640, int wHeight = 480,
                     bool wFullscreen = false, bool wResizable = true, bool wVsync = false);

    /// The destructor deallocates and cleans up any memory used by the RWindow program.
    /// @note If the window handle is not already destroyed, it will be done here.
    virtual ~RWindow();

    static RWindow* GetExtant() { return extant; };

    /// Bindables are provided for hooking into window instances conveniently, without the need to derive and override for simple use cases.
    Event<> OnOpenEvent;
    Event<> OnClosingEvent;
    Event<RWindowEvent> OnFocusLostEvent;
    Event<RWindowEvent> OnFocusGainEvent;
    Event<float> OnRefreshEvent;
    Event<WindowResizeRequestEvent> OnResizeRequestEvent;
    Event<KeyDownEvent> OnKeyDownEvent;
    Event<KeyUpEvent> OnKeyUpEvent;
    Event<MouseMoveEvent> OnMouseMoveEvent;
    Event<MouseButtonDownEvent> OnMouseButtonDownEvent;
    Event<MouseButtonUpEvent> OnMouseButtonUpEvent;
    Event<MouseWheelEvent> OnMouseWheelEvent;

    /// These methods can also be overridden in derived classes.
    /// Called upon the window requesting to open.
    virtual void OnOpen() {}
    /// Called right before the window closes.
    virtual void OnClosing() {}
    /// Called when the window loses focus.
    virtual void OnFocusLost(const RWindowEvent& e) {}
    /// Called when the window gains focus.
    virtual void OnFocusGain(const RWindowEvent& e) {}
    /// Called when the window is 'refreshed', in other words, a render pass is completed.
    virtual void OnRefresh(float elapsed) {}
    /// Called when a resize request has succeeded.
    virtual void OnResizeSuccess() {}
    /// Called when a resize request is sent to the operating system.
    virtual bool OnResizeRequest(const WindowResizeRequestEvent& e) { return true;}
    virtual void OnKeyDown(const KeyDownEvent&) {}
    virtual void OnKeyUp(const KeyUpEvent&) {}
    virtual void OnMouseMove(const MouseMoveEvent&) {}
    virtual void OnMouseButtonDown(const MouseButtonDownEvent&) {}
    virtual void OnMouseButtonUp(const MouseButtonUpEvent&) {}
    virtual void OnMouseWheel(const MouseWheelEvent&) {}

    /// Returns an IPair representing mouse coordinates relative to the top-left corner of the window.
    /// This result is cached from the operating-system, and as such may be out-of-date.
    /// @see GetAccurateMouseCoordinates().
    [[nodiscard]] IPair GetMouseCoordinates() const;
    [[nodiscard]] int GetMouseWheelPersistent() const { return currentMouse.Wheel; }
    [[nodiscard]] bool IsOpen() const { return open; }
    [[nodiscard]] bool IsClosing() const { return closing; }
    /// Returns whether the window currently has mouse and/or keyboard focus.
    [[nodiscard]] bool IsFocused() const;
    /// Returns whether the window is currently in Fullscreen.
    // TODO: Support Fullscreen, FullscreenWindowed, and Windowed?
    [[nodiscard]] bool IsFullscreen() const;
    /// Returns whether the window can be resized.
    [[nodiscard]] bool IsResizable() const;
    /// Returns whether V-Sync is enabled.
    [[nodiscard]] bool IsVsyncEnabled() const;
    /// Returns whether the window is considered to be alive. Once dead, any logic loop should be terminated, and the cleanup procedure should run.
    [[nodiscard]] bool IsAlive() const;
    /// Returns whether the given key is currently being pressed.
    [[nodiscard]] bool IsKeyDown(Key key) const;
    /// Returns whether the given mouse button is currently being pressed.
    [[nodiscard]] bool IsMouseButtonDown(const MouseButton &button) const;

    [[nodiscard]] float GetDeltaTime() const;
    /// Returns the approximate frames-per-second using delta time.
    [[nodiscard]] float GetRefreshRate() const;
    /// Returns the number of frames ran since the windows' creation.
    [[nodiscard]] float GetRefreshCounter() const;

    /// Cleans up and closes the window object.
    void Close();
    /// Closes the window immediately, potentially without allowing finalization to occur.
    void ForceClose();
    void ForceCloseAndTerminateProgram();

    /** Display a small window with some text and an "OK" button.
     * I tried un-defining the macro, Calling it in a lambda,
     * Calling MessageBoxA, Putting the call to the MessageBox
     * function in its own namespace and then calling that.
     * It just wouldn't let me name it MessageBox - Redacted. */

    /// @note Execution of the parent window is stopped while the message box is up.
    void UniqueFunctionNameForMessageBoxBecauseMicrosoftUsesMacros(const std::string& title, const std::string& message);

    /// Sets whether or not to make the window fullscreen.
    /// @note This is implemented per-OS, and as such, it simply requests the OS to do what we want. No guarantee about follow-through can be given.
    void SetFullscreen(bool fs);

    /// Sets whether or not to make the window resizable.
    /// @note This is implemented per-OS, and as such, it simply requests the OS to do what we want. No guarantee about follow-through can be given.
    void SetResizable(bool resizable);

    /// Sets the title of this window.
    void SetTitle(const std::string& title);
    [[nodiscard]] std::string GetTitle() const;

    /// Returns the horizontal length of this window, in pixels.
    [[nodiscard]] int GetWidth() const;
    /// Returns the vertical length of this window, in pixels.
    [[nodiscard]] int GetHeight() const;

    /// This is unfortunately here because of the poor design of windows. Maybe once interfaces are done this won't be required anymore.
    void SetSizeWithoutEvent(const IPair& size);
    void SetLastKnownWindowSize(const IPair& size);

    /// Tells the underlying window manager to destroy this window and drop the handle.
    /// The window, in theory, can not be re-opened after this.
    void DestroyOSWindowHandle();

    /// Reads events from the operating system, and processes them accordingly.
    /// TODO: Move out of public API, consumers should call Refresh or ideally an update() call.
    void PollEvents();

    /// Updates internal window state, similar to ManagedRefresh, but without accounting for any timekeeping. This is left up to the user.
    void Refresh();

    /// Updates the window and handles timing internally.
    void ManagedRefresh();

    /// Requests the operating system to change the window size.
    /// @param width
    /// @param height
    void SetSize(int width, int height);
    /// Requests the operating system to change the window size.
    /// @param size
    void SetSize(const IPair& size);

    /// Returns the position of the window's top-left corner relative to the display
    IPair GetPos() const;

    /// Returns the known size of the window, in {x,y} pixel measurement.
    IPair GetSize() const;

    /// Returns the position of the "renderable area" of the window relative to it's top left corner.
    /// (used to account for the width or the border & title bar).
    IPair GetPositionOfRenderableArea() const;

    /// Requests the operating system to move the window to the specified coordinates on the display.
    /// @param x The horizontal screen position to place the window at.
    /// @param y The vertical screen position to place the window at.
    void SetPos(int x, int y);
    /// Requests the operating system to move the window to the specified coordinates on the display.
    /// @param pos A IPair representing the x,y coordinates of the desired window destination. Fractional values are ignored.
    void SetPos(const IPair& pos);

    /// Pull the window to the top, such that it is displayed on top of everything else.
    /// NOTE: The implementation is defined per-OS, and thus there is no guarantee of it always working.
    void Raise();

    /// Push the window Lower, such that it is effectively hidden behind other windows.
    /// NOTE: The implementation is defined per-OS, and thus there is no guarantee of it always working.
    void Lower();

    void SetCursorStyle(CursorStyle style) const;

    void SetCursorCustomIcon() const;

    void SetCursorLocked();

    void SetCursorCenter();

    void RestoreCursorFromLastCenter(); // Feels nicer for users

    /// Hides the cursor when it's inside of our window. Useful for 3D game camera.
    void SetCursorVisible(bool cursor_enable);

    bool GetCursorVisible();

    /// Returns the current time, represented as a high-resolution std::chrono alias.
    static std::chrono::steady_clock::time_point GetTimestamp();

    /// Computes elapsed time from a start-point and end-point.
    float ComputeElapsedFrameTimeSeconds(std::chrono::steady_clock::time_point start, std::chrono::steady_clock::time_point end);

    /// Updates internals to account for the latest calculated frame time.
    void UpdateFrameTiming(float frame_time);
public:
/// These unfortunately *have* to be public because of the poor design of the windows event loop.
    void processKeyRelease (Key key);
    void processKeyPress (Key key);
    /// @note This will be invoked **before** the window-close procedure begins.
    void processOnClose();
    /// @note This will be invoked **after** the window-open procedure completes.
    void processOnOpen();
    void processMousePress(const MouseButton& btn);
    void processMouseRelease(const MouseButton& btn);
    void processFocusIn();
    void processFocusOut();
    void processMouseMove(IPair last_pos, IPair new_pos);
    void processMouseWheel(int scrolls);

    /// Virtual functions which *must* be overridden based on the Renderer.
public:
    virtual void SwapBuffers() = 0;

    /// Sets whether or not to enable vertical synchronization.
    /// @note This is implemented per-OS, and as such, it simply requests the OS to do what we want. No guarantee about follow-through can be given.
    virtual void SetVsyncEnabled(bool state) = 0;

    /// Returns the name of the developer of the user's graphics driver, if it can be determined.
    virtual std::string GetGraphicsDriverVendor() = 0;

    /// This function instructs the operating system to create the actual window, and give it to us to control.
    /// Calling this function, therefore, creates the real 'window' object on the operating system.
    [[nodiscard]] virtual bool Open() = 0;

    /// Bring attention to the user, On X11 this flashes the entry for this window in the window list.
    /// @note If the window is already in focus when this is called nothing happens.
    void Flash();

    /// @returns True if we are definitely running on a software renderer.
    /// @note For some APIs this isn't typically possible.
    [[nodiscard]] virtual bool SoftwareRendered() { return false; }
};

// TODO in the event that we can't find OpenGL or the Open() fails, have a way to say so without throwing an exception.
class ReWindow::OpenGLWindow : public RWindow {
protected:
    uint8_t gl_major, gl_minor;
public:
    void SwapBuffers() override;
    void SetVsyncEnabled(bool state) override;
    std::string GetGraphicsDriverVendor() override;
    [[nodiscard]] bool Open() override;
    [[nodiscard]] bool SoftwareRendered() override;
public:
    OpenGLWindow(const std::string& title, int width, int height, uint8_t gl_major, uint8_t gl_minor);
};

class ReWindow::VulkanWindow : public RWindow {
public:
    void SwapBuffers() override;
    void SetVsyncEnabled(bool state) override;
    std::string GetGraphicsDriverVendor() override;
    [[nodiscard]] bool Open() override;
    VulkanWindow(const std::string& title, int width, int height);
};