#include <ReWindow/types/Window.h>
#include <ReWindow/Logger.h>
#include <GL/glx.h>
#include <dlfcn.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>

using namespace ReWindow;
class ReWindow::RWindow::Platform {
public:
    Window window;
    XEvent xev;
    Display* display;
    int defaultScreen;
    XVisualInfo* visual;
    XSetWindowAttributes xSetWindowAttributes;
    XWindowAttributes windowAttributes;
    Atom wmDeleteWindow;
    Atom windowTypeAtom;
    Atom windowTypeUtilityAtom;
    XSizeHints hints;
    Cursor invisible_cursor = 0;
    XWMHints* wm_hints = nullptr;
};

void* glx_lib = nullptr;
void* opengl_lib = nullptr;

namespace OpenGL {
    bool constructor_success = false;

    typedef const GLubyte *(*glGetString_t)(GLenum name);
    glGetString_t glGetString = nullptr;

    typedef Bool (*glXMakeCurrent_t)(Display *dpy, GLXDrawable drawable, GLXContext ctx);
    glXMakeCurrent_t glXMakeCurrent = nullptr;

    typedef void *(*glXGetProcAddressARB_t)(const GLubyte *procname);
    glXGetProcAddressARB_t glXGetProcAddressARB = nullptr;

    typedef void (*glXSwapBuffers_t)(Display *dpy, GLXDrawable drawable);
    glXSwapBuffers_t glXSwapBuffers = nullptr;

    typedef GLXContext (*glXCreateContext_t)(Display *dpy, XVisualInfo *vis, GLXContext shareList, Bool direct);
    glXCreateContext_t glXCreateContext = nullptr;

    typedef GLXFBConfig *(*glXChooseFBConfig_t)(Display *dpy, int screen, const int *attribList, int *nItems);
    glXChooseFBConfig_t glXChooseFBConfig = nullptr;

    typedef XVisualInfo *(*glXGetVisualFromFBConfig_t)(Display *dpy, GLXFBConfig config);
    glXGetVisualFromFBConfig_t glXGetVisualFromFBConfig = nullptr;

    typedef XVisualInfo *(*glXChooseVisual_t)(Display *dpy, int screen, const int *attribList);
    glXChooseVisual_t glXChooseVisual = nullptr;

    typedef GLXContext (*glXCreateContextAttribsARBProc)(Display *, GLXFBConfig, GLXContext, Bool, const int *);
}

GLXContext gl_context;


void OpenGLWindow::SwapBuffers() {
    OpenGL::glXSwapBuffers(platform->display,platform->window);
}

void OpenGLWindow::SetVsyncEnabled(bool state) {
    PFNGLXSWAPINTERVALEXTPROC glXSwapIntervalEXT;
    glXSwapIntervalEXT = (PFNGLXSWAPINTERVALEXTPROC) OpenGL::glXGetProcAddressARB((const GLubyte *) "glXSwapIntervalEXT");
    glXSwapIntervalEXT(platform->display, platform->window, state);
}

std::string OpenGLWindow::GetGraphicsDriverVendor() {
    return std::string((const char*) OpenGL::glGetString(GL_VENDOR));
}

bool OpenGLWindow::Open() {
    if (!OpenGL::constructor_success)
        return false;

    platform->display = XOpenDisplay(nullptr);
    platform->defaultScreen = DefaultScreen(platform->display);

    platform->xSetWindowAttributes.border_pixel = BlackPixel(platform->display, platform->defaultScreen);
    platform->xSetWindowAttributes.background_pixel = BlackPixel(platform->display, platform->defaultScreen);
    platform->xSetWindowAttributes.override_redirect = True;
    platform->xSetWindowAttributes.event_mask = ExposureMask;

    auto glXCreateContextAttribsARB = (OpenGL::glXCreateContextAttribsARBProc) OpenGL::glXGetProcAddressARB((const GLubyte*)"glXCreateContextAttribsARB");
    XVisualInfo* vi = nullptr;

    // Fallback to the old way if you somehow don't have this.
    if (!glXCreateContextAttribsARB) {
        GLint glAttributes[] = {GLX_RGBA, GLX_DEPTH_SIZE, 24, GLX_DOUBLEBUFFER, None};
        platform->visual = OpenGL::glXChooseVisual(platform->display, platform->defaultScreen, glAttributes);
        gl_context = OpenGL::glXCreateContext(platform->display, platform->visual, nullptr, GL_TRUE);
        platform->window = XCreateWindow(platform->display, RootWindow(platform->display, platform->defaultScreen), 0, 0, width, height, 0, platform->visual->depth,
                                     InputOutput, platform->visual->visual, CWBackPixel | CWColormap | CWBorderPixel | NoEventMask, &platform->xSetWindowAttributes);
        Logger::Debug("Created OpenGL Context with glXCreateContext.");

        if (!platform->window)
            return false;
    }

    else {
        static int visual_attributes[]
                {
                        GLX_X_RENDERABLE, True, GLX_DRAWABLE_TYPE, GLX_WINDOW_BIT,
                        GLX_RENDER_TYPE, GLX_RGBA_BIT, GLX_X_VISUAL_TYPE, GLX_TRUE_COLOR,
                        GLX_RED_SIZE, 8, GLX_GREEN_SIZE, 8, GLX_BLUE_SIZE, 8,
                        GLX_ALPHA_SIZE, 8, GLX_DEPTH_SIZE, 24, GLX_STENCIL_SIZE, 8,
                        GLX_DOUBLEBUFFER, True, None
                };

        int fb_count;
        GLXFBConfig* fb_configurations = OpenGL::glXChooseFBConfig(platform->display, platform->defaultScreen, visual_attributes, &fb_count);
        if (!fb_configurations || fb_count == 0)
            return false;

        GLXFBConfig best_fbc = fb_configurations[0];
        vi = OpenGL::glXGetVisualFromFBConfig(platform->display, best_fbc);

        if (!vi)
            return false;

        platform->xSetWindowAttributes.colormap = XCreateColormap(platform->display, RootWindow(platform->display, vi->screen), vi->visual, AllocNone);
        platform->window = XCreateWindow(platform->display, RootWindow(platform->display, vi->screen), 0, 0, width, height, 0, vi->depth, InputOutput,
                                     vi->visual, CWBackPixel | CWColormap | CWBorderPixel, &platform->xSetWindowAttributes);

        // TODO allow the user to specify what OpenGL version they want.
        int context_attributes[] { GLX_CONTEXT_MAJOR_VERSION_ARB, gl_major, GLX_CONTEXT_MINOR_VERSION_ARB, gl_minor, None };
        gl_context = glXCreateContextAttribsARB(platform->display, best_fbc, nullptr, True, context_attributes);
        XFree(fb_configurations);
        Logger::Debug("Created OpenGL Context with glXCreateContextAttribsARB");
    }

    if (!gl_context)
        return false;
    if (!OpenGL::glXMakeCurrent(platform->display, platform->window, gl_context))
        return false;

    if (vi)
        XFree(vi);

    XSelectInput(platform->display, platform->window, ExposureMask | KeyPressMask | KeyReleaseMask | ButtonPressMask | ButtonReleaseMask | PointerMotionMask |
    PointerMotionHintMask | FocusChangeMask | StructureNotifyMask | SubstructureRedirectMask | SubstructureNotifyMask | CWColormap );

    XMapWindow(platform->display, platform->window);
    XStoreName(platform->display, platform->window, title.c_str());

    platform->windowTypeAtom = XInternAtom(platform->display, "_NET_WM_WINDOW_TYPE", False);
    platform->windowTypeUtilityAtom = XInternAtom(platform->display, "_NET_WM_WINDOW_TYPE_UTILITY", False);
    XChangeProperty(platform->display, platform->window, platform->windowTypeAtom, XA_ATOM, 32, PropModeReplace, (unsigned char*) &platform->windowTypeUtilityAtom, 1);

    platform->wmDeleteWindow = XInternAtom(platform->display, "WM_DELETE_WINDOW", False);
    XSetWMProtocols(platform->display, platform->window, &platform->wmDeleteWindow, 1);

    XGetWindowAttributes(platform->display, platform->window, &platform->windowAttributes);
    render_area_position = { platform->windowAttributes.x, platform->windowAttributes.y };

    open = true;
    processOnOpen();
    return true;
}

OpenGLWindow::OpenGLWindow(const std::string& title, int width, int height, uint8_t gl_major, uint8_t gl_minor)
: gl_major(gl_major), gl_minor(gl_minor), RWindow(title, width, height) {

    // Try this first.
    glx_lib = dlopen("libGLX.so", RTLD_NOLOAD);
    opengl_lib = dlopen("libOpenGL.so", RTLD_NOLOAD);

    if (!glx_lib)
        glx_lib = dlopen("libGLX.so", RTLD_LAZY);
    if (!opengl_lib)
        opengl_lib = dlopen("libOpenGL.so", RTLD_LAZY);

    if (!opengl_lib)
        Logger::Error("libOpenGL.so couldn't be found in your LD_LIBRARY_PATH."),
        OpenGL::constructor_success = false;

    if (!glx_lib)
        Logger::Error("libGLX.so couldn't be found in your LD_LIBRARY_PATH."),
        OpenGL::constructor_success = false;

    OpenGL::glGetString = (OpenGL::glGetString_t) dlsym(opengl_lib, "glGetString");
    OpenGL::glXMakeCurrent = (OpenGL::glXMakeCurrent_t) dlsym(glx_lib, "glXMakeCurrent");
    OpenGL::glXGetProcAddressARB = (OpenGL::glXGetProcAddressARB_t) dlsym(glx_lib, "glXGetProcAddressARB");
    OpenGL::glXSwapBuffers = (OpenGL::glXSwapBuffers_t) dlsym(glx_lib, "glXSwapBuffers");
    OpenGL::glXCreateContext = (OpenGL::glXCreateContext_t) dlsym(glx_lib, "glXCreateContext");
    OpenGL::glXChooseFBConfig = (OpenGL::glXChooseFBConfig_t) dlsym(glx_lib, "glXChooseFBConfig");
    OpenGL::glXGetVisualFromFBConfig = (OpenGL::glXGetVisualFromFBConfig_t) dlsym(glx_lib, "glXGetVisualFromFBConfig");
    OpenGL::glXChooseVisual = (OpenGL::glXChooseVisual_t) dlsym(glx_lib, "glXChooseVisual");
    OpenGL::constructor_success = true;
}

bool OpenGLWindow::SoftwareRendered() {
    std::string renderer((const char*) OpenGL::glGetString(GL_RENDERER));
    if (renderer.find("llvmpipe"))
        return true;
    if (renderer.find("softpipe"))
        return true;
    return false;
}
