#include <ReWindow/types/Window.h>
#include <ReWindow/data/WindowsEventLoop.h>
#include <ReWindow/Logger.h>
#include <GL/gl.h>

class RWindow::Platform {
public:
    HINSTANCE hInstance;
    HWND hwnd;
    HDC hdc;
};

using namespace ReWindow;

HGLRC glContext;
HMODULE gl_lib = nullptr;
bool constructor_success = true;

namespace OpenGL {
    typedef const GLubyte* (APIENTRY *GLGETSTRINGPROC)(GLenum name);
    GLGETSTRINGPROC glGetString = nullptr;

    typedef BOOL (WINAPI *WGLMAKECURRENTPROC)(HDC hdc, HGLRC hglrc);
    WGLMAKECURRENTPROC wglMakeCurrent = nullptr;

    typedef FARPROC (WINAPI *WGLGETPROCADDRPROC)(LPCSTR lpszProc);
    WGLGETPROCADDRPROC wglGetProcAddress = nullptr;


    typedef HGLRC (WINAPI *WGLCREATECONTEXTPROC)(HDC hdc);
    WGLCREATECONTEXTPROC wglCreateContext = nullptr;

}
bool OpenGLWindow::Open() {
    if (!constructor_success)
        return false;

  platform->hInstance = GetModuleHandle(nullptr);

  WNDCLASS wc = { };
  wc.lpfnWndProc = WindowProc;
  wc.hInstance = platform->hInstance;
  wc.lpszClassName = "RWindowClass";
  RegisterClass(&wc);
  platform->hwnd = CreateWindowEx(
      0,
      "RWindowClass",
      title.c_str(),
      WS_OVERLAPPEDWINDOW,
      CW_USEDEFAULT, CW_USEDEFAULT, width, height,
      nullptr,
      nullptr,
      platform->hInstance,
      nullptr
  );
    SetWindowLongPtr(platform->hwnd, GWLP_USERDATA, reinterpret_cast<LONG_PTR>(this));
  PIXELFORMATDESCRIPTOR pfd = {
      sizeof(PIXELFORMATDESCRIPTOR),
      1,
      PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
      PFD_TYPE_RGBA,
      24,
      0, 0, 0, 0, 0, 0,
      0,
      0,
      0,
      0, 0, 0, 0,
      24,
      8,
      0,
      PFD_MAIN_PLANE,
      0,
      0, 0, 0
  };

  platform->hdc = GetDC(platform->hwnd);
  int pixelFormat = ChoosePixelFormat(platform->hdc, &pfd);
  SetPixelFormat(platform->hdc, pixelFormat, &pfd);
  glContext = OpenGL::wglCreateContext(platform->hdc);
  OpenGL::wglMakeCurrent(platform->hdc, glContext);
  ShowWindow(platform->hwnd, SW_SHOW);
  open = true;
    return true;
}

std::string OpenGLWindow::GetGraphicsDriverVendor() {
    return std::string(reinterpret_cast<const char*>(OpenGL::glGetString(GL_VENDOR)));
}

void OpenGLWindow::SwapBuffers() {
    ::SwapBuffers(platform->hdc);
}

void OpenGLWindow::SetVsyncEnabled(bool b) {
    typedef BOOL(WINAPI* PFNWGLSWAPINTERVALEXTPROC)(int interval);
    auto wglSwapIntervalEXT = (PFNWGLSWAPINTERVALEXTPROC) OpenGL::wglGetProcAddress("wglSwapIntervalEXT");

    if (wglSwapIntervalEXT)
        wglSwapIntervalEXT(b ? 1 : 0);
}

bool OpenGLWindow::SoftwareRendered() {
    std::string renderer(reinterpret_cast<const char*>(OpenGL::glGetString(GL_RENDERER)));
    if (renderer.find("llvmpipe"))
        return true;
    if (renderer.find("softpipe"))
        return true;
    if (renderer.find("GDI Generic"))
        return true;
    if (GetGraphicsDriverVendor().find("Mesa"))
        return true;
    return false;
}

OpenGLWindow::OpenGLWindow(const std::string& title, int width, int height, uint8_t gl_major, uint8_t gl_minor)
: gl_major(gl_major), gl_minor(gl_minor), RWindow(title, width, height) {
    gl_lib = LoadLibrary("opengl32.dll");

    if (!gl_lib)
        Logger::Error("opengl32.dll couldn't be found. If you have it, you can try putting it in the same directory as the executable."),
        constructor_success = false;

    OpenGL::glGetString = (OpenGL::GLGETSTRINGPROC) GetProcAddress(gl_lib, "glGetString");
    OpenGL::wglMakeCurrent = (OpenGL::WGLMAKECURRENTPROC) GetProcAddress(gl_lib, "wglMakeCurrent");
    OpenGL::wglGetProcAddress = (OpenGL::WGLGETPROCADDRPROC) GetProcAddress(gl_lib, "wglGetProcAddress");
    OpenGL::wglCreateContext = (OpenGL::WGLCREATECONTEXTPROC) GetProcAddress(gl_lib, "wglCreateContext");
}

